/*
btopen.c Version 0.1.0  Btree open database
Copyright (C) 2003  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

#include <bthash.h>

#define BTOPEN 200

btreefmt *btopen(char *dbname, int rwflag,
   int *reason)
   {
   DWORD msgerr;
   int   rslt;
   DWORD rdlen;
   HANDLE hndl;
   unsigned char *p,*q;
   btreefmt *bthdr;
   LPVOID lpMsgBuf;
   *reason = 0;
   if (rwflag == UPDT)
      {
      hndl = CreateFile(dbname,
         GENERIC_READ | GENERIC_WRITE,
         0,              /* do not share */
         NULL,           /* cannot be inherited */
         OPEN_EXISTING,
         FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
         NULL);          /* no template */
      } /* if update */
   else if (rwflag == READ)
      {
      hndl = CreateFile(dbname,
         GENERIC_READ,
         0,              /* do not share */
         NULL,           /* cannot be inherited */
         OPEN_EXISTING,
         FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
         NULL);          /* no template */
      } /* if read only */
   else
      {
      fprintf(stderr,"btopen: invalid read/write flag\n");
      *reason = BTREE_ERR + BTOPEN + 1;
      return(NULL);
      } /* rwflag err */
   if (hndl == INVALID_HANDLE_VALUE)
      {
      msgerr = GetLastError();
      *reason = (int) SYS_ERR + msgerr;
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,             /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,                /* message length */
         NULL);            /* insertion values */
      fprintf(stderr,"btopen: %s", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"btopen: %s: open ", dbname);
      if (rwflag == READ)
         fprintf(stderr,"read-only\n");
      else if (rwflag == UPDT)
         fprintf(stderr,"read/write\n");
      return(NULL);
      } /* if open error */
   bthdr = (btreefmt *) malloc(ROOT+ROOT);
   if (bthdr == NULL)
      {
      fprintf(stderr,"btopen: out of memory "
	 "allocating bthdr\n");
      rslt = CloseHandle(hndl);
      *reason = BTREE_ERR + BTOPEN + 3;
      return(NULL);
      } /* out of mem */
   rslt = ReadFile(hndl,bthdr,ROOT,
      &rdlen,NULL);
   if (!rslt)
      {
      msgerr = GetLastError();
      *reason = (int) SYS_ERR + msgerr;
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,             /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,                /* message length */
         NULL);            /* insertion values */
      fprintf(stderr,"btopen: %s", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"btopen: dbname %s\n", dbname);
      fprintf(stderr,"btopen: error "
         "reading header\n");
      rslt = CloseHandle(hndl);
      free(bthdr);
      return(NULL);
      } /* read err */
   if (rdlen != ROOT)
      {
      fprintf(stderr,"btopen: dbname %s\n", dbname);
      fprintf(stderr,"btopen: error "
         "reading header\n");
      rslt = CloseHandle(hndl);
      free(bthdr);
      *reason = BTREE_ERR + BTOPEN + 4;
      return(NULL);
      } /* read err */
   if (strncmp(bthdr->signature,
      BT_SIGNATURE,sizeof(BT_SIGNATURE)))
      {
      fprintf(stderr,"btopen: dbname %s\n", dbname);
      fprintf(stderr,"btopen: error "
         "opening non-btree file\n");
      rslt = CloseHandle(hndl);
      free(bthdr);
      *reason = BTREE_ERR + BTOPEN + 5;
      return(NULL);
      } /* invalid signature */
   p = (unsigned char *) bthdr + ROOT;
   q = (unsigned char *) p + ROOT;
   while (p < q) *p++ = '\0';
   bthdr->hndl = hndl;

   /* nodesz = records per node */
   bthdr->rcdnode = bthdr->rcdhdr + bthdr->rcdlen;
   bthdr->blknode = bthdr->blkhdr + bthdr->blksz;
   bthdr->midofst = bthdr->blkhdr
      + (bthdr->rcdnode * (bthdr->nodesz >> 1));

   bthdr->head = (stkfmt *) btstkini(bthdr);
   bthdr->tail = (stkfmt *) btstkini(bthdr);
   bthdr->head->prev = bthdr->head;
   bthdr->head->next = bthdr->tail;
   bthdr->tail->prev = bthdr->head;
   bthdr->tail->next = bthdr->tail;
   bthdr->blk = (blkfmt *) malloc(bthdr->blksz+16);
   if (bthdr->blk == NULL)
      {
      fprintf(stderr,"btopen: out of memory "
         "allocating bthdr->buf\n");
      btclose(bthdr);
      free(bthdr);
      *reason = BTREE_ERR + BTOPEN + 6;
      return(NULL);
      } /* out of mem */
   bthdr->blkleft = (blkfmt *) malloc(bthdr->blksz+16);
   if (bthdr->blkleft == NULL)
      {
      fprintf(stderr,"btopen: out of memory "
	 "allocating bthdr->buf2\n");
      btclose(bthdr);
      free(bthdr->blk);
      free(bthdr);
      *reason = BTREE_ERR + BTOPEN + 7;
      return(NULL);
      } /* out of mem */
   p = (unsigned char *) bthdr->blkleft;
   q = (unsigned char *) p + bthdr->blksz;
   while (p < q) *p++ = '\0';
   bthdr->blkrght = (blkfmt *) malloc(bthdr->blksz+16);
   if (bthdr->blkrght == NULL)
      {
      fprintf(stderr,"btopen: out of memory "
	 "allocating bthdr->buf3\n");
      btclose(bthdr);
      free(bthdr->blk);
      free(bthdr->blkleft);
      free(bthdr);
      *reason = BTREE_ERR + BTOPEN + 8;
      return(NULL);
      } /* out of mem */
   p = (unsigned char *) bthdr->blkrght;
   q = (unsigned char *) p + bthdr->blksz;
   while (p < q) *p++ = '\0';
   bthdr->node = (bt_nodefmt *) malloc(bthdr->rcdnode+16);
   if (bthdr->node == NULL)
      {
      fprintf(stderr,"btopen: out of memory "
	 "allocating bthdr->node\n");
      btclose(bthdr);
      free(bthdr->blk);
      free(bthdr->blkleft);
      free(bthdr->blkrght);
      free(bthdr);
      *reason = BTREE_ERR + BTOPEN + 9;
      return(NULL);
      } /* out of mem */
   p = (unsigned char *) bthdr->node;
   q = (unsigned char *) p + bthdr->rcdnode;
   while (p < q) *p++ = '\0';
   strcpy(bthdr->dbname,dbname);
   bthdr->rwflag = rwflag;
   return(bthdr);
   } /* btopen */
