/*
btreorg.c Version 0.1.0  Btree reorganize database
Copyright (C) 2003  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

#include <bthash.h>

int btreorg(btreefmt *bthdr)
   {
   BOOL   bResult;
   DWORD  msgerr;
   LPVOID lpMsgBuf;
   int rslt;
   int position;
   int opnerr;
   unsigned char *rndSeed;
   unsigned char *p,*q;
   btreefmt *tmpbt;
   int rwflag;
   unsigned char dbname[256];
   unsigned char tmpbtdbn[32];
   unsigned char tmpdbname[32];
   unsigned char buf[4096 * 16];
   bthdr->reason = 0;
   tmpbt = (btreefmt *) btopent(bthdr->nodesz,
      bthdr->rcdlen,
      bthdr->keylen,
      bthdr->keyofst,
      bthdr->dletofst,
      &opnerr,NULL);
   if (tmpbt == NULL)
      {
      fprintf(stderr,"btreorg: error initializing "
	 "the temporary database: %d\n",
	 opnerr);
      bthdr->reason = BTREE_ERR + 2601;
      return(BTREE_ERR + 2601);
      } /* can't create temp bt */
   position = SWPFRST;
   rslt = 0;
   while (!rslt)
      {
      rslt = btsweep(bthdr,position,buf);
      if (!rslt)
	 {
         p = (unsigned char *) bthdr->rcd;
         if (*(p + bthdr->dletofst) == DELETED)
            {
            position = SWPNEXT;
            continue;
            } /* if deleted */
	 rslt = btisrt(tmpbt,ROOT,buf);
	 if (rslt)
	    {
            fprintf(stderr,"btreorg: error inserting in "
	       "the temporary database: %d reason %d\n",
	       rslt, bthdr->reason);
	    btcloset(tmpbt);
            bthdr->reason = BTREE_ERR + 2602;
            return(BTREE_ERR + 2602);
	    } /* if insert err */
	 } /* if good read */
      position = SWPNEXT;
      } /* read loop */
   if (bthdr->reason)
      {
      fprintf(stderr,"btreorg: btsweep error %d\n",
	 bthdr->reason);
      btcloset(tmpbt);
      return(BTREE_ERR + 2603);
      } /* btsweep err */
   rndSeed = (unsigned char *) rndinit();
   p = (unsigned char *) tmpdbname;
   q = (unsigned char *) p + 8;
   while (p < q)
      {
      *p++ = rndnum(26,rndSeed) + 'A';
      } /* for each char in temp dbname */
   *p = '\0';
   strcat((char *) tmpdbname,".BT");
   strcpy((char *) dbname,bthdr->dbname);
   strcpy((char *) tmpbtdbn,tmpbt->dbname);
   btclose(tmpbt);
   rwflag = bthdr->rwflag;
   btclose(bthdr);
   bResult = MoveFile((char *) dbname,
      (char *) tmpdbname);
   if (!bResult)
      {
      msgerr = GetLastError();
      bthdr->reason = SYS_ERR + msgerr;
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,             /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,                /* message length */
         NULL);            /* insertion values */
      fprintf(stderr,"btreorg: %s", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"btreorg: error renaming %s to %s\n",
	 bthdr->dbname, tmpdbname);
      return(SYS_ERR + msgerr);
      } /* rename failed */
   bResult = MoveFile((char *) tmpbtdbn,
      (char *) dbname);
   if (!bResult)
      {
      msgerr = GetLastError();
      bthdr->reason = SYS_ERR + msgerr;
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,             /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,                /* message length */
         NULL);            /* insertion values */
      fprintf(stderr,"btreorg: %s", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"btreorg: error renaming %s to %s\n",
	 tmpbt->dbname, bthdr->dbname);
      return(SYS_ERR + msgerr);
      } /* rename failed */
   bResult = DeleteFile((char *) tmpdbname);
   if (!bResult)
      {
      msgerr = GetLastError();
      bthdr->reason = SYS_ERR + msgerr;
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,             /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,                /* message length */
         NULL);            /* insertion values */
      fprintf(stderr,"btreorg: %s", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"btreorg: error removing %s\n",
	 tmpdbname);
      return(SYS_ERR + msgerr);
      } /* remove failed */

   /* The database is closed. */
   /* The calling program has to call btopen */
   /* before using the database again. */

   return(0);
   } /* btreorg */
