/*
hinit.c - Release 0.1.0  Hash initialize database
Copyright (C) 2003  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

#include <bthash.h>

#define RCDHDR        4
#define MAXRCDS 9999999
#define MAXRCDLEN 65536

int HClsInit(HANDLE hndl)
   {
   BOOL   bResult;
   DWORD  msgerr;
   LPVOID lpMsgBuf;
   bResult = CloseHandle(hndl);
   if (!bResult)
      {
      msgerr = GetLastError();
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,           /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,              /* msglen */
         NULL);          /* insertion values */
      fprintf(stderr,"HClsInit: %s\n", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"HClsInit: close "
         "error %d\n", msgerr);
      return(SYS_ERR + msgerr);
      } /* close err */
   return(0);
   } /* HClsInit */

int HGetBlkInit(HANDLE hndl, int rcdlen,
   unsigned char *rcd)
   {
   BOOL   bResult;
   DWORD  msgerr;
   LPVOID lpMsgBuf;
   DWORD  rdlen;
   bResult = ReadFile(hndl,rcd,
      (DWORD) rcdlen,
      &rdlen,NULL);
   if (!bResult)
      {
      msgerr = GetLastError();
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,           /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,              /* msglen */
         NULL);          /* insertion values */
      fprintf(stderr,"HGetBlkInit: %s\n", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"HGetBlkInit: read "
         "error %d\n", msgerr);
      HClsInit(hndl);
      return(SYS_ERR + msgerr);
      } /* read err */
   if (rdlen == (DWORD) rcdlen)
      return(0);
   else
      {
      fprintf(stderr,"HGetBlkInit: read error "
	 "(rdlen %d)\n", rdlen);
      HClsInit(hndl);
      return(HASH_ERR + 107);
      } /* read error */
   } /* HGetBlkInit */

int HPutBlkInit(HANDLE hndl, int rcdlen,
   unsigned char *rcd)
   {
   BOOL   bResult;
   DWORD  msgerr;
   LPVOID lpMsgBuf;
   DWORD  wrtlen;
   bResult = WriteFile(hndl,rcd,
      (DWORD) rcdlen,
      &wrtlen,NULL);
   if (!bResult)
      {
      msgerr = GetLastError();
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,           /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,              /* msglen */
         NULL);          /* insertion values */
      fprintf(stderr,"HPutBlkInit: %s\n", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"HPutBlkInit: write "
         "error %d\n", msgerr);
      HClsInit(hndl);
      return(SYS_ERR + msgerr);
      } /* write err */
   if (wrtlen == (DWORD) rcdlen)
      return(0);
   else
      {
      fprintf(stderr,"HPutBlkInit: write error "
	 "(wrtlen %d)", wrtlen);
      HClsInit(hndl);
      return(HASH_ERR + 108);
      } /* write error */
   return(0);
   } /* HPutBlkInit */

int HBldHdrInit(HANDLE hndl, int totrcds,
   int rcdlen, int keylen, int keyofst)
   {
   int i;
   int rslt;
   int hdrsz;
   int nodelen;
   h_nodefmt *node;
   hashfmt *blk;
   unsigned char *p,*q;
   rslt = 0;
   hdrsz = HDRLEN;
   blk = (hashfmt *) malloc(hdrsz+16);
   if (blk == NULL)
      {
      fprintf(stderr,"HBldHdrInit: malloc error\n");
      HClsInit(hndl);
      return(HASH_ERR + 110);
      } /* malloc err */
   p = (unsigned char *) blk;
   q = (unsigned char *) blk + hdrsz;
   while (p < q) *p++ = '\0';
   strcpy(blk->signature,H_SIGNATURE);
   blk->totrcds = totrcds;
   blk->rcdlen  = rcdlen;
   blk->keylen  = keylen;
   blk->keyofst = keyofst;
   blk->rcdhdr  = RCDHDR;
   blk->hdrlen  = HDRLEN;
   nodelen = rcdlen + RCDHDR;
   rslt = HPutBlkInit(hndl,HDRLEN,
      (unsigned char *) blk);
   if (rslt) return(rslt);
   /* blank records for hashing area */
   p = (unsigned char *) blk;
   q = (unsigned char *) blk + nodelen;
   while (p < q) *p++ = '\0';
   node = (h_nodefmt *) blk;
   node->next = UINT_MAX;
   i = totrcds;
   while (i--)
      {
      rslt = HPutBlkInit(hndl,nodelen,
         (unsigned char *) node);
      if (rslt) break;
      } /* initialize database */
   free(blk);
   return(rslt);
   } /* HBldHdrInit */

int HTruncInitOld(char *dbname,HANDLE *hndl)
   {
   DWORD  msgerr;
   LPVOID lpMsgBuf;
   *hndl = CreateFile(dbname,
      GENERIC_READ | GENERIC_WRITE,
      0,           /* do not share */
      NULL,        /* cannot be inherited */
      TRUNCATE_EXISTING,
      FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
      NULL);       /* no template */
   if (*hndl == INVALID_HANDLE_VALUE)
      {
      msgerr = GetLastError();
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,             /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,                /* message length */
         NULL);            /* insertion values */
      fprintf(stderr,"HTruncInitOld: %s\n", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"HTruncInitOld: %s: "
	 "open read/write\n",
         dbname);
      fprintf(stderr,"HTruncInitOld: "
         "CreateFile error\n");
      return(SYS_ERR + msgerr);
      } /* if open error */
   return(0);
   } /* HTruncInitOld */

int HOpnInitOld(char *dbname,HANDLE *hndl)
   {
   DWORD  msgerr;
   DWORD  rslt;
   HANDLE oldhndl;
   unsigned char buf[32];
   LPVOID lpMsgBuf;
   *hndl = NULL;
   oldhndl = CreateFile(dbname,
      GENERIC_READ,
      0,           /* do not share */
      NULL,        /* cannot be inherited */
      OPEN_EXISTING,
      FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
      NULL);       /* no template */
   if (oldhndl == INVALID_HANDLE_VALUE)
      {
      msgerr = GetLastError();
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,             /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,                /* message length */
         NULL);            /* insertion values */
      fprintf(stderr,"HOpnInitOld: %s\n", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"HOpnInitOld: %s: "
	 "open error\n",
         dbname);
      return(SYS_ERR + msgerr);
      } /* open err */
   rslt = HGetBlkInit(oldhndl,16,buf);
   if (rslt) return(rslt);
   rslt = strncmp((char *)buf,H_SIGNATURE,16); 
   if (rslt)
      {
      fprintf(stderr,"HOpnInitOld: filename %s "
         "for non-hash file\n",
         dbname);
      HClsInit(oldhndl);
      return(HASH_ERR + 106);
      } /* not hash file */
   else
      {
      rslt = HClsInit(oldhndl);
      if (rslt) return(rslt);
      } /* close after good signature */
   return(HTruncInitOld(dbname,hndl));
   } /* HOpnInitOld */

int HOpnInitNew(char *dbname, HANDLE *hndl)
   {
   DWORD  msgerr;
   LPVOID lpMsgBuf;
   *hndl = CreateFile(dbname,
      GENERIC_READ | GENERIC_WRITE,
      0,           /* do not share */
      NULL,        /* cannot be inherited */
      CREATE_NEW,
      FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
      NULL);       /* no template */
   if (*hndl == INVALID_HANDLE_VALUE)
      {
      msgerr = (DWORD) GetLastError();
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,             /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,                /* message length */
         NULL);            /* insertion values */
      fprintf(stderr,"HOpnInitNew: %s\n", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"HOpnInitNew: %s: open read/write\n",
         dbname);
      if (msgerr == ERROR_FILE_EXISTS)
         {
         fprintf(stderr,"HOpnInitNew: %s: file exists\n",
            dbname);
         } /* file exists */
      return(SYS_ERR + msgerr);
      }
   return(0);
   } /* HOpnInitNew */

int HOpnInitAll(char *dbname, HANDLE *hndl)
   {
   int rslt;
   DWORD  msgerr;
   HANDLE oldhndl;
   unsigned char buf[32];
   LPVOID lpMsgBuf;
   *hndl = NULL;
   oldhndl = CreateFile(dbname,
      GENERIC_READ,
      0,           /* do not share */
      NULL,        /* cannot be inherited */
      OPEN_EXISTING,
      FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
      NULL);       /* no template */
   if (oldhndl != INVALID_HANDLE_VALUE)
      {
      rslt = HGetBlkInit(oldhndl,16,buf);
      if (rslt) return(rslt);
      rslt = strncmp((char *)buf,H_SIGNATURE,16); 
      if (rslt)
	 {
         fprintf(stderr,"HOpnInitAll: filename %s "
	    "for non-hash file\n",
	    dbname);
         HClsInit(oldhndl);
         return(HASH_ERR + 115);
	 } /* not hash file */
      else
	 {
         rslt = HClsInit(oldhndl);
	 if (rslt) return(rslt);
         return(HTruncInitOld(dbname,hndl));
	 } /* close after good signature */
      } /* if good open rdonly */
   fprintf(stderr,"Creating %s\n", dbname);
   *hndl = CreateFile(dbname,
      GENERIC_READ | GENERIC_WRITE,
      0,           /* do not share */
      NULL,        /* cannot be inherited */
      CREATE_NEW,
      FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
      NULL);       /* no template */
   if (*hndl == INVALID_HANDLE_VALUE)
      {
      msgerr = GetLastError();
      FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS,
         NULL,             /* source */
         msgerr,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
         (LPTSTR) &lpMsgBuf,
         0,                /* message length */
         NULL);            /* insertion values */
      fprintf(stderr,"HOpnInitAll: %s\n", lpMsgBuf);
      LocalFree(lpMsgBuf);
      fprintf(stderr,"HOpnInitAll: %s: "
         "open read/write\n",
         dbname);
      return(SYS_ERR + msgerr);
      }
   return(0);
   } /* HOpnInitAll */

int HOpnInit(char *dbname, int truncflag,
   int totrcds, int rcdlen, int keylen,
   int keyofst)
   {
   int rslt;
   HANDLE hndl;
   if (truncflag == INITOLD)
      rslt = HOpnInitOld(dbname,&hndl);
   else if (truncflag == INITNEW)
      rslt = HOpnInitNew(dbname,&hndl);
   else if (truncflag == INITALL)
      rslt = HOpnInitAll(dbname,&hndl);
   if (rslt) return(rslt);

   rslt = HBldHdrInit(hndl,totrcds,
      rcdlen,keylen,keyofst);
   if (rslt) return(rslt);
   return(HClsInit(hndl));
   } /* HOpnInit */

int HValidateInit(char *dbname, int truncflag,
   int totrcds, int rcdlen,
   int keylen, int keyofst)
   {
   if (keylen > rcdlen)
      {
      fprintf(stderr,"hinit: key-length %d "
         "exceeds record-length %d\n",
	 keylen, rcdlen);
      return(HASH_ERR + 112);
      }
   if (keyofst > rcdlen - keylen)
      {
      fprintf(stderr,"hinit: key-offset %d "
         "exceeds the record-length %d\n",
	 keyofst, rcdlen);
      fprintf(stderr,"    minus the key-length %d\n",
	 keylen);
      return(HASH_ERR + 113);
      }
   return(HOpnInit(dbname,truncflag,
      totrcds,rcdlen,keylen,keyofst));
   } /* HValidateInit */

int hinit(char *dbname, int truncflag,
   int totrcds, int rcdlen,
   int keylen, int keyofst)
   {
   int namelen;
   namelen = strlen(dbname);
   if (namelen < 1 || namelen > 128)
      {
      fprintf(stderr,"hinit: wrong "
         "length dbname %s\n",
	 dbname);
      return(HASH_ERR + 116);
      } /* dbname wrong length */
   if (truncflag == INITNEW
      || truncflag == INITOLD
      || truncflag == INITALL);
   else
      {
      fprintf(stderr,"hinit: invalid "
	 "truncate flag %d\n",
	 truncflag);
      return(HASH_ERR + 114);
      } /* bad truncflag */
   if (totrcds < 1 || totrcds > MAXRCDS)
      {
      fprintf(stderr,"hinit: invalid "
         "totrcds %d\n",
         totrcds);
      return(HASH_ERR + 101);
      }
   if (totrcds != hprime(totrcds))
      {
      fprintf(stderr,"hinit: totrcds "
         "is not prime\n");
      return(HASH_ERR + 117);
      }
   if (rcdlen < 1 || rcdlen > MAXRCDLEN)
      {
      fprintf(stderr,"hinit: invalid "
         "record-length %d\n",
         rcdlen);
      return(HASH_ERR + 103);
      }
   if (keylen < 1 || keylen > MAXRCDLEN)
      {
      fprintf(stderr,"hinit: invalid "
         "key-length %d\n",
         keylen);
      return(HASH_ERR + 104);
      }
   if (keyofst < 0 || keyofst > MAXRCDLEN-1)
      {
      fprintf(stderr,"hinit: invalid "
         "key-offset %d\n",
         keyofst);
      return(HASH_ERR + 105);
      }
   return(HValidateInit(dbname,truncflag,
      totrcds,rcdlen,keylen,keyofst));
   } /* hinit */
